/* The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Mobile Application Link.
 *
 * The Initial Developer of the Original Code is AvantGo, Inc.
 * Portions created by AvantGo, Inc. are Copyright (C) 1997-1999
 * AvantGo, Inc. All Rights Reserved.
 *
 */

/* syncmal.c
 *
 * J-Pilot plugin callback functions by Jason Day, jasonday@worldnet.att.net
 * http://jasonday.home.att.net
 */




#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <utime.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <dirent.h>
#include <dlfcn.h>

#include <pi-dlp.h>
#include <pi-file.h>

#include "syncmal.h"
#include "libplugin.h"
#include "plugin_prefs.h"





#include <AGUtil.h>
#include <AGClientProcessor.h>
#include <AGProtocol.h>
#include <AGBufferReader.h>
#include <AGPalmProtocol.h>
#include <AGUserConfig.h>
#include <AGServerConfig.h>
#include <AGSyncCommon.h>
#include <AGCommandProcessor.h>
#include <AGDesktopInfoPalm.h>
#include <AGTimeoutMessageBox.h>
#include <AGMobileLinkSettings.h>


#include <pi-source.h>
#include <pi-socket.h>
#include <pi-version.h>

#define DEFAULT_CARD_NUM 0

static char *RCSID = "$Id: syncmal.c,v 1.10 2000/03/11 18:16:14 jday Exp $";
#define VERSION "0.30"

/*
static gboolean glob_enabled = TRUE;
static gboolean glob_proxy_enabled = FALSE;
static gboolean glob_socks_enabled = FALSE;
*/

#define SMPREF_ENABLED 0
#define SMPREF_USE_PROXY 1
#define SMPREF_PROXY_ADDRESS 2
#define SMPREF_PROXY_PORT 3
#define SMPREF_PROXY_USERNAME 4
#define SMPREF_PROXY_PASSWORD 5
#define SMPREF_USE_SOCKS 6
#define SMPREF_SOCKS_ADDRESS 7
#define SMPREF_SOCKS_PORT 8

#define NUM_SMPREFS 9
prefType syncmal_prefs[NUM_SMPREFS] = {
    {"enabled", INTTYPE, INTTYPE, 1, ""},
    {"use_proxy", INTTYPE, INTTYPE, 0, ""},
    {"proxy_address", CHARTYPE, CHARTYPE, 0, ""},
    {"proxy_port", CHARTYPE, CHARTYPE, 0, ""},
    {"proxy_username", CHARTYPE, CHARTYPE, 0, ""},
    {"proxy_password", CHARTYPE, CHARTYPE, 0, ""},
    {"use_socks", INTTYPE, INTTYPE, 0, ""},
    {"socks_address", CHARTYPE, CHARTYPE, 0, ""},
    {"socks_port", CHARTYPE, CHARTYPE, 0, ""}
};

/*
 * These two arrays just hold pointers to some of the labels and text
 * entries. This makes it easier to enable/disable them and cuts down
 * on the number of global variables.
 */
#define PROXY_WIDGETS 8
static GtkWidget *proxy_widgets[PROXY_WIDGETS];
#define SOCKS_WIDGETS 4
static GtkWidget *socks_widgets[SOCKS_WIDGETS];

static void cb_save (GtkWidget *widget, gpointer data);
static void cb_enabled (GtkWidget *widget, gpointer data);
static void cb_proxy_enabled (GtkWidget *widget, gpointer data);
static void cb_socks_enabled (GtkWidget *widget, gpointer data);
static void cb_proxy_address (GtkWidget *widget, GtkWidget *entry);
static void cb_proxy_port (GtkWidget *widget, GtkWidget *entry);
static void cb_proxy_username (GtkWidget *widget, GtkWidget *entry);
static void cb_proxy_password (GtkWidget *widget, GtkWidget *entry);
static void cb_socks_address (GtkWidget *widget, GtkWidget *entry);
static void cb_socks_port (GtkWidget *widget, GtkWidget *entry);

static void Connect();
static char *httpProxy      = NULL;
static int   httpProxyPort  = 0;
static char *socksProxy     = NULL;
static int   socksProxyPort = 0;
static char *proxyUsername  = NULL;
static char *proxyPassword  = NULL;
static char *progname       = NULL;

#ifdef _DEBUG
#define debug_print(x) printf("%s\n", (x));
#else
#define debug_print(x)
#endif

#ifdef AP_SNPRINTF
#define snprintf ap_snprintf
#define vsnprintf ap_vsnprintf
#endif

typedef struct {

    AGDeviceInfo * deviceInfo;
    AGUserConfig * userConfig;
    AGServerConfig * currentServerConfig;
    AGClientProcessor * clientProcessor;
    AGPlatformCalls * platform;
    AGRecord * record;
    AGDBConfig * currentDb;
    AGCommandProcessor *commandProcessor;
    AGBool quit;

    /* Pilot-specific */
    AGBool currentDBIsResourceType;
    int pilot_rHandle;
    int pilot_RecIndex;
    recordid_t id;
    uint8  *pilot_buffer;
    int pilot_buffer_size;

    /* Secure Network Library Stuff */
    AGBool          hasseclib;

} PalmSyncInfo;


typedef sword (*netInitFunc)(AGNetCtx *ctx);
typedef sword (*netCloseFunc)(AGNetCtx *ctx);
typedef int32 (*netCtxSizeFunc)(void);
typedef void  (*netPreSyncHook) (AGNetCtx *ctx,
                                 AGServerConfig *sc,
                                 AGLocationConfig *lc,
                                 AGSyncProcessor *sp,
                                 AGBool connectSecure);
typedef void  (*netPostSyncHook) (AGNetCtx *ctx,
                                  AGServerConfig *sc,
                                  AGLocationConfig *lc,
                                  AGSyncProcessor *sp,
                                  AGBool connectSecure);

static netInitFunc     secnetinit = NULL;
static netCloseFunc    secnetclose = NULL;
static netCtxSizeFunc  secctxsize = NULL;
static netPreSyncHook  secnetpresync = NULL;
static netPostSyncHook  secnetpostsync = NULL;
static int sd = 0;
static int verbose = 1;

static AGBool setupPlatformCalls(PalmSyncInfo * pInfo);
PalmSyncInfo * syncInfoNew();
void syncInfoFree(PalmSyncInfo * pInfo);
static AGUserConfig *getUserConfig(uint32 * pilotID);
static void storeDeviceUserConfig(AGUserConfig *userConfig, recordid_t id);
static AGBool doClientProcessorLoop(PalmSyncInfo * pInfo, AGNetCtx *ctx);
static int32 initAndOpenDatabase(void *out, AGDBConfig *theDatabase,
                                 int32 *errCode);
static int32 getNextModifiedRecord(void * out, AGRecord ** record,
                                   int32 *errCode);
static int32 getNextRecord(void * out, AGRecord ** record,
                           int32 *errCode);

static int32 cmdTASK(void *out, int32 *returnErrorCode, char *currentTask,
                     AGBool bufferable);
static int32 cmdITEM(void *out, int32 *returnErrorCode,
                     int32 currentItemNumber,
                     int32 totalItemCount,
                     char *currentItem);
static int32 cmdDELETEDATABASE(void *out, int32 *returnErrorCode,
                               char *dbname);
static int32 cmdOPENDATABASE(void *out, int32 *returnErrorCode,
                             char *dbname);
static int32 cmdCLOSEDATABASE(void *out, int32 *returnErrorCode);
static int32 cmdCLEARMODS(void *out, int32 *returnErrorCode);
static int32 cmdGOODBYE(void *out, int32 *returnErrorCode,
                        AGSyncStatus syncStatus,
                        int32 errorCode,
                        char *errorMessage);
static int32 cmdRECORD(void *out, int32 *returnErrorCode,
                       int32 *newUID,
                       int32 uid,
                       AGRecordStatus mod,
                       int32 recordDataLength,
                       void *recordData,
                       int32 platformDataLength,
                       void *platformData);






int plugin_get_name(char *name, int len)
{
    strncpy (name,
             "SyncMAL "
             VERSION,
             len);
    return 0;
}

int plugin_get_menu_name(char *name, int len)
{
    strncpy(name, "SyncMAL", len);
    return 0;
}

int plugin_get_help_name(char *name, int len)
{
    strncpy(name, "About SyncMAL", len);
    return 0;
}

int plugin_help(char **text, int *width, int *height)
{
    /* We could also pass back *text=NULL */
    *text = strdup(
                   /*-------------------------------------------*/
                   "SyncMAL plugin for J-Pilot\r\n"
                   "version " VERSION "\r\n"
                   "Jason Day (c) 2000.\r\n"
                   "jasonday@worldnet.att.net\r\n"
                   "http://jasonday.home.att.net\r\n"
                  );
    *height = 200;
    *width = 300;
    return 0;
}

int plugin_startup (jp_startup_info *info)
{
    int i;
    jp_init();
    jp_logf (LOG_DEBUG, "SyncMAL: plugin_startup\n");

    /* Open the prefs file */
    if (jp_read_rc_file ("syncmal.rc", syncmal_prefs, NUM_SMPREFS) < 0) {
        /* Most likely means the rc file hasn't been created yet */
    }
    else {
        jp_logf (LOG_DEBUG, "SyncMAL: loaded preferences from syncmal.rc\n");
    }
    /*
    for (i = 0; i < NUM_SMPREFS; i++) {
        jp_logf (LOG_DEBUG, "prefs %d name = %s\n", i, syncmal_prefs[i].name);
    }
    */
    return 0;
}

int plugin_sync (int sd_hack) {
    PalmSyncInfo * pInfo;
    uint32 pilotID;
    AGNetCtx *ctx;
    long ivalue;
    const char *svalue;

    jp_logf (LOG_DEBUG, "SyncMAL: plugin_sync\n");

    jp_get_pref (syncmal_prefs, SMPREF_ENABLED, &ivalue, &svalue);
    if (ivalue == 0) {
        jp_logf (LOG_INFO | LOG_GUI, "SyncMAL: skipping at user request\n");
        return 0;
    }

    pInfo = syncInfoNew();
    if (NULL == pInfo)
        return -1;

    sd = sd_hack;

    if (loadSecLib()) {
        ctx = calloc(1, (*secctxsize)());
        (*secnetinit)(ctx);
    } else {
        ctx = calloc( 1, sizeof( AGNetCtx));
        AGNetInit( ctx);
    }

    if (setupPlatformCalls(pInfo))
        return -1;

    pInfo->userConfig = getUserConfig(&pilotID);

    doClientProcessorLoop(pInfo, ctx);

    storeDeviceUserConfig(pInfo->userConfig, pilotID);

    if (secnetclose) {
        (*secnetclose)(ctx);
    } else {
        AGNetClose(ctx);
    }

    syncInfoFree(pInfo);
    free(ctx);

return 0;

}

/*
 * This function is called by J-Pilot when the user selects this plugin
 * from the plugin menu, or from the search window when a search result
 * record is chosen.  In the latter case, unique ID will be set.  This
 * application should go directly to that record in the case.
 */
int plugin_gui (GtkWidget *vbox, GtkWidget *hbox, unsigned int unique_id)
{
    GtkWidget *vbox1;
    GtkWidget *button;
    GtkWidget *temp_hbox;
    GtkWidget *label;
    GtkWidget *entry;
    GtkWidget *table;
    long ivalue;
    const char *svalue;

    jp_logf(LOG_DEBUG, "SyncMAL: plugin gui started\n");

    vbox1 = gtk_vbox_new (FALSE, 0);
    gtk_box_pack_start (GTK_BOX(hbox), vbox1, TRUE, FALSE, 0);

    /* ----------------------------------- */
    /* Top portion of screen (preferences) */
    /* ----------------------------------- */

    button = gtk_check_button_new_with_label ("Enabled");
    gtk_signal_connect (GTK_OBJECT (button), "toggled",
                        GTK_SIGNAL_FUNC (cb_enabled), NULL);

    jp_get_pref (syncmal_prefs, SMPREF_ENABLED, &ivalue, &svalue);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), ivalue);
    gtk_box_pack_start (GTK_BOX(vbox1), button, FALSE, FALSE, 5);

    gtk_box_pack_start (GTK_BOX(vbox1), gtk_hseparator_new(), FALSE, FALSE, 0);

    /* --------------*/
    /* Proxy Options */
    /* --------------*/
    button = gtk_check_button_new_with_label ("Use Proxy");
    jp_get_pref (syncmal_prefs, SMPREF_USE_PROXY, &ivalue, &svalue);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), ivalue);
    gtk_signal_connect (GTK_OBJECT (button), "toggled",
                        GTK_SIGNAL_FUNC (cb_proxy_enabled), NULL);
    gtk_box_pack_start (GTK_BOX(vbox1), button, FALSE, FALSE, 5);


    table = gtk_table_new (4, 2, FALSE);

    label = gtk_label_new ("Proxy Address:");
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    proxy_widgets[0] = label;
    gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 0, 1);
    entry = gtk_entry_new();
    proxy_widgets[1] = entry;
    jp_get_pref (syncmal_prefs, SMPREF_PROXY_ADDRESS, &ivalue, &svalue);
    if (svalue) {
        gtk_entry_set_text (GTK_ENTRY(entry), svalue);
    }
    gtk_signal_connect(GTK_OBJECT(entry), "changed",
                       GTK_SIGNAL_FUNC(cb_proxy_address),
                       entry);
    gtk_table_attach (GTK_TABLE (table), entry, 1, 2, 0, 1,
                      (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                      (GtkAttachOptions) (0), 10, 0);


    label = gtk_label_new ("Proxy Port:");
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    proxy_widgets[2] = label;
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 1, 2);
    entry = gtk_entry_new();
    proxy_widgets[3] = entry;
    jp_get_pref (syncmal_prefs, SMPREF_PROXY_PORT, &ivalue, &svalue);
    if (svalue) {
        gtk_entry_set_text (GTK_ENTRY(entry), svalue);
    }
    gtk_signal_connect(GTK_OBJECT(entry), "changed",
                       GTK_SIGNAL_FUNC(cb_proxy_port),
                       entry);
    gtk_table_attach (GTK_TABLE (table), entry, 1, 2, 1, 2,
                      (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                      (GtkAttachOptions) (0), 10, 0);


    label = gtk_label_new ("Proxy Username:");
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    proxy_widgets[4] = label;
    gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 2, 3);
    entry = gtk_entry_new();
    proxy_widgets[5] = entry;
    jp_get_pref (syncmal_prefs, SMPREF_PROXY_USERNAME, &ivalue, &svalue);
    if (svalue) {
        gtk_entry_set_text (GTK_ENTRY(entry), svalue);
    }
    gtk_signal_connect(GTK_OBJECT(entry), "changed",
                       GTK_SIGNAL_FUNC(cb_proxy_username),
                       entry);
    gtk_table_attach (GTK_TABLE (table), entry, 1, 2, 2, 3,
                      (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                      (GtkAttachOptions) (0), 10, 0);


    label = gtk_label_new ("Proxy Password:");
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    proxy_widgets[6] = label;
    gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 3, 4);
    entry = gtk_entry_new();
    proxy_widgets[7] = entry;
    jp_get_pref (syncmal_prefs, SMPREF_PROXY_PASSWORD, &ivalue, &svalue);
    if (svalue) {
        gtk_entry_set_text (GTK_ENTRY(entry), svalue);
    }
    gtk_signal_connect(GTK_OBJECT(entry), "changed",
                       GTK_SIGNAL_FUNC(cb_proxy_password),
                       entry);
    gtk_entry_set_visibility (GTK_ENTRY(entry), FALSE);
    gtk_table_attach (GTK_TABLE (table), entry, 1, 2, 3, 4,
                      (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                      (GtkAttachOptions) (0), 10, 0);


    gtk_box_pack_start (GTK_BOX(vbox1), table, FALSE, FALSE, 0);
    cb_proxy_enabled (button, NULL);

    gtk_box_pack_start (GTK_BOX(vbox1), gtk_hseparator_new(), FALSE, FALSE, 5);

    /* --------------*/
    /* Socks Options */
    /* --------------*/
    button = gtk_check_button_new_with_label ("Use SOCKS");
    jp_get_pref (syncmal_prefs, SMPREF_USE_SOCKS, &ivalue, &svalue);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), ivalue);
    gtk_signal_connect (GTK_OBJECT (button), "toggled",
                        GTK_SIGNAL_FUNC (cb_socks_enabled), NULL);
    gtk_box_pack_start (GTK_BOX(vbox1), button, FALSE, FALSE, 5);

    table = gtk_table_new (2, 2, FALSE);

    label = gtk_label_new ("SOCKS Proxy:");
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    socks_widgets[0] = label;
    gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 0, 1);
    entry = gtk_entry_new();
    socks_widgets[1] = entry;
    jp_get_pref (syncmal_prefs, SMPREF_SOCKS_ADDRESS, &ivalue, &svalue);
    if (svalue) {
        gtk_entry_set_text (GTK_ENTRY(entry), svalue);
    }
    gtk_signal_connect(GTK_OBJECT(entry), "changed",
                       GTK_SIGNAL_FUNC(cb_socks_address),
                       entry);
    gtk_table_attach (GTK_TABLE (table), entry, 1, 2, 0, 1,
                      (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                      (GtkAttachOptions) (0), 10, 0);

    label = gtk_label_new ("SOCKS Port:");
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    socks_widgets[2] = label;
    gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 1, 2);
    entry = gtk_entry_new();
    socks_widgets[3] = entry;
    jp_get_pref (syncmal_prefs, SMPREF_SOCKS_PORT, &ivalue, &svalue);
    if (svalue) {
        gtk_entry_set_text (GTK_ENTRY(entry), svalue);
    }
    gtk_signal_connect(GTK_OBJECT(entry), "changed",
                       GTK_SIGNAL_FUNC(cb_socks_port),
                       entry);
    gtk_table_attach (GTK_TABLE (table), entry, 1, 2, 1, 2,
                      (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                      (GtkAttachOptions) (0), 10, 0);

    gtk_box_pack_start (GTK_BOX(vbox1), table, FALSE, FALSE, 5);
    cb_socks_enabled (button, NULL);

    /* ------------ */
    /* Apply button */
    /* ------------ */
    gtk_box_pack_start (GTK_BOX(vbox1), gtk_hseparator_new(), FALSE, FALSE, 5);

    button = gtk_button_new_with_label ("Save Changes");
    gtk_signal_connect (GTK_OBJECT(button), "clicked",
                        GTK_SIGNAL_FUNC(cb_save), NULL);
    gtk_box_pack_start (GTK_BOX(vbox1), button, FALSE, FALSE, 5);


    gtk_widget_show_all (hbox);
    return 0;
}

/* ----------------------------------------------------------------- */

static void cb_save (GtkWidget *widget, gpointer data) {
    jp_write_rc_file ("syncmal.rc", syncmal_prefs, NUM_SMPREFS);
}

static void cb_enabled (GtkWidget *widget, gpointer data)
{
    gboolean enabled = GTK_TOGGLE_BUTTON(widget)->active;
    jp_set_pref_int (syncmal_prefs, SMPREF_ENABLED, enabled);
}

static void cb_proxy_enabled (GtkWidget *widget, gpointer data)
{
    int i;
    gboolean enabled = GTK_TOGGLE_BUTTON(widget)->active;

    jp_set_pref_int (syncmal_prefs, SMPREF_USE_PROXY, enabled);
    for (i = 0; i < PROXY_WIDGETS; i++) {
        gtk_widget_set_sensitive (proxy_widgets[i], enabled);
    }
}

static void cb_socks_enabled (GtkWidget *widget, gpointer data)
{
    int i;
    gboolean enabled = GTK_TOGGLE_BUTTON(widget)->active;

    jp_set_pref_int (syncmal_prefs, SMPREF_USE_SOCKS, enabled);
    for (i = 0; i < SOCKS_WIDGETS; i++) {
        gtk_widget_set_sensitive (socks_widgets[i], enabled);
    }
}

static void cb_proxy_address (GtkWidget *widget, GtkWidget *entry)
{
    httpProxy = gtk_entry_get_text(GTK_ENTRY(entry));
    jp_set_pref_char (syncmal_prefs, SMPREF_PROXY_ADDRESS, httpProxy);
}

static void cb_proxy_port (GtkWidget *widget, GtkWidget *entry)
{
    httpProxyPort = atoi (gtk_entry_get_text(GTK_ENTRY(entry)));
    jp_set_pref (syncmal_prefs,
                 SMPREF_PROXY_PORT,
                 httpProxyPort,
                 gtk_entry_get_text(GTK_ENTRY(entry)));
}

static void cb_proxy_username (GtkWidget *widget, GtkWidget *entry)
{
    proxyUsername = gtk_entry_get_text(GTK_ENTRY(entry));
    jp_set_pref_char (syncmal_prefs, SMPREF_PROXY_USERNAME, proxyUsername);
}

static void cb_proxy_password (GtkWidget *widget, GtkWidget *entry)
{
    proxyPassword = gtk_entry_get_text(GTK_ENTRY(entry));
    jp_set_pref_char (syncmal_prefs, SMPREF_PROXY_PASSWORD, proxyPassword);
}

static void cb_socks_address (GtkWidget *widget, GtkWidget *entry)
{
    socksProxy = gtk_entry_get_text(GTK_ENTRY(entry));
}

static void cb_socks_port (GtkWidget *widget, GtkWidget *entry)
{
    socksProxyPort = atoi (gtk_entry_get_text(GTK_ENTRY(entry)));
    jp_set_pref (syncmal_prefs,
                 SMPREF_SOCKS_PORT,
                 socksProxyPort,
                 gtk_entry_get_text (GTK_ENTRY(entry)));
}




/* ----------------------------------------------------------------------------*/
void syncInfoFree(PalmSyncInfo * pInfo)
{

    if (NULL != pInfo) {

        if (NULL != pInfo->platform)
            free(pInfo->platform);

        if (NULL != pInfo->userConfig)
            AGUserConfigFree(pInfo->userConfig);

        if (NULL != pInfo->pilot_buffer)
            free(pInfo->pilot_buffer);

        if (NULL != pInfo->commandProcessor)
            AGCommandProcessorFree(pInfo->commandProcessor);

        free(pInfo);
    }
}

/* ----------------------------------------------------------------------------*/
PalmSyncInfo * syncInfoNew()
{
    PalmSyncInfo * pInfo;

    /* Allocate the PalmSyncInfo structure. */
    pInfo = (PalmSyncInfo *)malloc(sizeof(PalmSyncInfo));
    if (NULL != pInfo) {

        const int pbs = 65535;

        bzero(pInfo, sizeof(PalmSyncInfo));

        pInfo->pilot_buffer_size    = pbs;
        pInfo->pilot_buffer         = (uint8 *)malloc(pbs);
        if (NULL == pInfo->pilot_buffer)
            goto fail;

        /* Allocate the platform calls record. */
        pInfo->platform = (AGPlatformCalls *)malloc(sizeof(AGPlatformCalls));
        bzero(pInfo->platform, sizeof(AGPlatformCalls));
        if (NULL == pInfo->platform)
            goto fail;

        return pInfo;
    }

    fail:
    if (NULL != pInfo)
        syncInfoFree(pInfo);

    return NULL;
}
/* ----------------------------------------------------------------------------*/
static AGBool setupPlatformCalls(PalmSyncInfo * pInfo)
{
    pInfo->platform->out = pInfo;
    pInfo->platform->nextModifiedRecordFunc = getNextModifiedRecord;
    pInfo->platform->nextRecordFunc = getNextRecord;
    pInfo->platform->openDatabaseFunc =  initAndOpenDatabase;
    return FALSE;
}
/* ----------------------------------------------------------------------------*/
static int readDeviceInfo(PalmSyncInfo * pInfo)
{
    AGDeviceInfo * devInfo;
    struct SysInfo sysInfo;
    struct CardInfo cardInfo;
    int majorver, minorver;
    char osverstring[24];
    int err;

    /* Start with clean slate. */
    devInfo = pInfo->deviceInfo;

    err = dlp_ReadSysInfo(sd, &sysInfo);
    if (err < 0) {
        jp_logf (LOG_FATAL, "dlp_ReadSysInfo failed with err %d\n", err);
        return -1;
    }

    cardInfo.card = DEFAULT_CARD_NUM;
    err = dlp_ReadStorageInfo(sd, DEFAULT_CARD_NUM, &cardInfo);
    if (err < 0) {
        jp_logf (LOG_FATAL, "dlp_ReadStorageInfo failed with err %d\n", err);
        return -1;
    }

    /* Find out version of operating system. */
    majorver = sysInfo.romVersion >> 24;
    minorver = 0xff & (sysInfo.romVersion >> 16);
    if (minorver < 10)
        minorver = 0;

    snprintf(osverstring, 24, "%d.%d", majorver, minorver);

    devInfo->availableBytes   = cardInfo.ramFree;
    devInfo->serialNumber     = strdup("");
    devInfo->osName           = strdup("PALM_OS");
    devInfo->osVersion        = strdup(osverstring);
    devInfo->screenWidth      = 150;
    devInfo->screenHeight     = 150;
    if(majorver > 2)
        devInfo->colorDepth    = 2;
    else
        devInfo->colorDepth    = 1;

    return 0;
}
/* ----------------------------------------------------------------------------*/
static AGBool getPalmDatabaseCreationInfo(AGDBConfig * db, uint32 * creator,
                                          uint32 * flags, uint32 * type)
{
    AGBufferReader * r = NULL;
    debug_print("GetPalmDatabaseCreationInfo()\n");

    if (NULL == db)
        return FALSE;

    if (0 == db->platformDataLength || NULL == db->platformData)
        return FALSE;

    r = AGBufferReaderNew((uint8*)db->platformData);
    if (NULL == r)
        return FALSE;

    AGPalmReadDBConfigPlatformData((AGReader*)r, creator, type, flags);
    AGBufferReaderFree(r);
    return TRUE;
}

/* ----------------------------------------------------------------------------*/
static int createDatabase(AGDBConfig *db)
{
    int dbhandle;
    long creator;
    int flags;
    int cardNo = DEFAULT_CARD_NUM;
    long type;
    int version = 0;
    int err;

    if (!db)
        return 0;

#ifdef _DEBUG
    printf("createDatabase\n");
#endif

    getPalmDatabaseCreationInfo(db, (uint32*)&creator,
                                (uint32*)&flags, (uint32*)&type);

    if ((err = dlp_CreateDB(sd,  creator,  type, cardNo, flags,
                            version, db->dbname, &dbhandle)) < 0) {
#ifdef _DEBUG
        printf("createDatabase: dlp_CreateDB failed err = %d\n", err);
#endif

        dbhandle = 0;
    }

    return dbhandle;
}

/* ----------------------------------------------------------------------------*/
static long openDatabase(PalmSyncInfo *pInfo, char *dbname, AGBool create)
{

    long result;

    if (!dbname || !pInfo) {
        debug_print("\n");
        return -1;
    }
    debug_print("... opening '");
    debug_print(dbname);
    debug_print("' ... ");

    pInfo->currentDb =
        AGServerConfigGetDBConfigNamed(pInfo->currentServerConfig, dbname);

    result = dlp_OpenDB(sd, DEFAULT_CARD_NUM, dlpOpenRead|dlpOpenWrite,
                        dbname, &pInfo->pilot_rHandle);

    if ((result < 0) && create)
        pInfo->pilot_rHandle = createDatabase(pInfo->currentDb);

    if (pInfo->pilot_rHandle) {
        uint32 creator, flags, type;
        if (getPalmDatabaseCreationInfo(pInfo->currentDb, &creator, &flags,
                                        &type) && (flags & 0x01) != 0) {
            pInfo->currentDBIsResourceType = TRUE;
        } else
            pInfo->currentDBIsResourceType = FALSE;
        debug_print("successfully.\n");
    } else {
        debug_print("unsuccessfully.\n");
        pInfo->currentDBIsResourceType = FALSE;
        pInfo->currentDb = NULL;
    }

    return result;
}

/* ----------------------------------------------------------------------------*/
static long closeDatabase(PalmSyncInfo * pInfo)
{
    int result;

    result = dlp_CloseDB(sd, pInfo->pilot_rHandle);
    pInfo->pilot_rHandle = 0;
    pInfo->currentDb = NULL;
    pInfo->currentDBIsResourceType = FALSE;
    return result;
}

/* ----------------------------------------------------------------------------*/
static void deleteDatabase(char * dbname)
{
    debug_print("deleteDatabase()\n");
    dlp_DeleteDB(sd, DEFAULT_CARD_NUM, dbname);
}
/* ----------------------------------------------------------------------------*/
static void clearMods(long dbHandle)
{
    debug_print("clearMods()\n");
    dlp_CleanUpDatabase(sd, dbHandle);
    dlp_ResetSyncFlags(sd, dbHandle);

}

/* ----------------------------------------------------------------------------*/
static long openUserConfigDatabase(void)
{
    long result;
    int userConfigDBHandle = 0;

    result = dlp_OpenDB(sd, DEFAULT_CARD_NUM, dlpOpenRead|dlpOpenWrite,
                        DEVICE_USERCONFIG_DB_NAME,
                        &userConfigDBHandle);

    if (result < 0) {

        result = dlp_CreateDB(sd, DEVICE_USERCONFIG_DB_CREATOR,
                              DEVICE_USERCONFIG_DB_TYPE, DEFAULT_CARD_NUM,
                              0, 0, DEVICE_USERCONFIG_DB_NAME,
                              &userConfigDBHandle);
        if (result < 0) {
            jp_logf (LOG_FATAL, "Unable to create user Config Databage\n");
            userConfigDBHandle = 0;
        }

    }
    return userConfigDBHandle;
}
#define BUFFERSIZE 0xFFFF
/* ----------------------------------------------------------------------------*/
static int32 readDeviceUserConfig(int userConfigDBHandle,
                                  AGUserConfig **deviceUserConfig)
{
    recordid_t id;
    int bufferSize = BUFFERSIZE;
    int attr = 0;
    int cat  = 0;
    int rc;
    uint8 buffer[BUFFERSIZE];
    AGBufferReader * r = NULL;

    rc = dlp_ReadRecordByIndex(sd, userConfigDBHandle, 0, (void *)buffer,
                               &id, &bufferSize, &attr, &cat);

    if (rc < 0) {
#ifdef _DEBUG
        printf("readDeviceUserConfig: dlp_ReadRecordByIndex failed, err = %d\n",
               rc);
#endif
        return 0;
    }

    r = AGBufferReaderNew(buffer);
    if (r) {
        *deviceUserConfig = AGUserConfigNewAndReadData((AGReader*)r);
        AGBufferReaderFree(r);
        return id;
    } else
        return 0;
}

/* ----------------------------------------------------------------------------*/
static void writeDeviceUserConfig(int userConfigDBHandle,
                                  AGUserConfig * deviceUserConfig,
                                  recordid_t *recID)
{
    AGBufferWriter * w = NULL;
    w = AGBufferWriterNew(0);
    if (w) {
        long result;
        AGUserConfigWriteData(deviceUserConfig, (AGWriter*)w);
        result =  dlp_WriteRecord(sd, userConfigDBHandle, 0,
                                  *recID, 0, (void *)AGBufferWriterGetBuffer(w),
                                  AGBufferWriterGetBufferSize(w),
                                  recID);
        AGBufferWriterFree(w);
    }
}

/* ----------------------------------------------------------------------------*/
static AGUserConfig *getUserConfig(uint32 * pilotID)
{

    /* The device is the truth. There is no way to set these values on
     the desktop */
    AGUserConfig * deviceUserConfig = NULL;
    int userConfigDBHandle = 0;

    /* Get device's record. */
    userConfigDBHandle = openUserConfigDatabase();
    if (userConfigDBHandle) {

        /* Retrieve device's idea of current userConfig. */
        *pilotID = readDeviceUserConfig(userConfigDBHandle,
                                        &deviceUserConfig);

        /* Done with database for now, so close it. */
        dlp_CloseDB(sd, userConfigDBHandle);


    }
    return deviceUserConfig;
}

/* ----------------------------------------------------------------------------*/
static void storeDeviceUserConfig(AGUserConfig *userConfig, recordid_t id)
{
    int userConfigDBHandle = openUserConfigDatabase();
    if (0 != userConfigDBHandle) {
        writeDeviceUserConfig(userConfigDBHandle,
                              userConfig, &id);
        dlp_CloseDB(sd, userConfigDBHandle);
    }
}

/* ----------------------------------------------------------------------------*/
static void doStartServer(PalmSyncInfo * pInfo,
                          AGServerConfig *sc,
                          int32 *errCode)
{
    pInfo->currentServerConfig = sc;
    if(pInfo->commandProcessor) {
        AGCommandProcessorFree(pInfo->commandProcessor);
        pInfo->commandProcessor = NULL;
    }
    pInfo->commandProcessor = AGCommandProcessorNew(sc);
    pInfo->platform->performCommandOut = pInfo->commandProcessor;
    pInfo->platform->performCommandFunc =
        AGCommandProcessorGetPerformFunc(pInfo->commandProcessor);

    pInfo->commandProcessor->commands.out = pInfo;

    pInfo->commandProcessor->commands.performTaskFunc = cmdTASK;
    pInfo->commandProcessor->commands.performItemFunc = cmdITEM;
    pInfo->commandProcessor->commands.performDeleteDatabaseFunc = cmdDELETEDATABASE;
    pInfo->commandProcessor->commands.performOpenDatabaseFunc = cmdOPENDATABASE;
    pInfo->commandProcessor->commands.performCloseDatabaseFunc = cmdCLOSEDATABASE;
    pInfo->commandProcessor->commands.performClearModsFunc = cmdCLEARMODS;
    pInfo->commandProcessor->commands.performGoodbyeFunc = cmdGOODBYE;
    pInfo->commandProcessor->commands.performRecordFunc = cmdRECORD;

}

/* ----------------------------------------------------------------------------*/
static void doEndServer(PalmSyncInfo * pInfo,
                        int32 *errCode)
{
    pInfo->currentServerConfig = NULL;
    if(pInfo->commandProcessor) {
        AGCommandProcessorFree(pInfo->commandProcessor);
        pInfo->commandProcessor = NULL;
    }
}
/* ----------------------------------------------------------------------------*/
static int32 cmdTASK(void *out, int32 *returnErrorCode, char *currentTask,
                     AGBool bufferable)
{
    if (currentTask) {
        jp_logf (LOG_DEBUG, "currentTask: %s\n");
        jp_logf (LOG_INFO | LOG_GUI, "%s\n", currentTask);
    }
    return AGCLIENT_CONTINUE;
}
/* ----------------------------------------------------------------------------*/
static int32 cmdITEM(void *out, int32 *returnErrorCode,
                     int32 currentItemNumber,
                     int32 totalItemCount,
                     char *currentItem)
{
    jp_logf (LOG_INFO | LOG_GUI, ".");
    fflush(stdout);
    if (currentItemNumber == totalItemCount)
        jp_logf (LOG_INFO | LOG_GUI, "\n");

    return AGCLIENT_CONTINUE;
}
/* ----------------------------------------------------------------------------*/
static int32 cmdDELETEDATABASE(void *out, int32 *returnErrorCode,
                               char *dbname)
{
    debug_print("doCmdAG_DELETEDATABASE_CMD()\n");

    if (NULL != dbname) {
        debug_print("... trying to delete database ");
        debug_print(dbname);
        debug_print(" from device.\n");
        deleteDatabase(dbname);
    }
    return AGCLIENT_CONTINUE;
}
/*----------------------------------------------------------------------------*/
static int32 cmdOPENDATABASE(void *out, int32 *returnErrorCode, char *dbname)
{
    PalmSyncInfo *pInfo = (PalmSyncInfo *)out;
    debug_print("doCmdAG_OPENDATABASE_CMD()\n");

    if (NULL != dbname) {
        long result;
        /* We assign a result code here only for debugging.  It's ok for an
         open to fail here. */
        result = openDatabase(pInfo, dbname, TRUE);
    }
    return AGCLIENT_CONTINUE;
}
/*----------------------------------------------------------------------------*/
static int32 cmdCLOSEDATABASE(void *out, int32 *returnErrorCode)
{
    PalmSyncInfo *pInfo = (PalmSyncInfo *)out;
    debug_print("doCmdAG_CLOSEDATABASE_CMD()\n");
    closeDatabase(pInfo);
    return AGCLIENT_CONTINUE;
}
/*----------------------------------------------------------------------------*/
static int32 cmdCLEARMODS(void *out, int32 *returnErrorCode)
{
    PalmSyncInfo *pInfo = (PalmSyncInfo *)out;
    debug_print("doCmdAG_CLEARMODS_CMD()\n");
    clearMods(pInfo->pilot_rHandle);
    return AGCLIENT_CONTINUE;
}
/*----------------------------------------------------------------------------*/
static int32 cmdGOODBYE(void *out, int32 *returnErrorCode,
                        AGSyncStatus syncStatus,
                        int32 errorCode,
                        char *errorMessage)
{
    debug_print("doCmdAG_GOODBYE_CMD()\n");

    if (errorMessage)
        jp_logf (LOG_INFO | LOG_GUI, "%s\n", errorMessage);

    return AGCLIENT_CONTINUE;
}
/*----------------------------------------------------------------------------*/
int getIndexFromPlatformData(uint8 *platformData)
{
    int16 recIndex;
    AGBufferReader reader;

    if (!platformData)
        return 0;

    AGBufferReaderInit(&reader, platformData);
    AGPalmReadRecordPlatformData((AGReader *)&reader, &recIndex);
    AGBufferReaderFinalize(&reader);
    return (int)recIndex;
}
/*----------------------------------------------------------------------------*/
static int32 cmdRECORD(void *out, int32 *returnErrorCode,
                       int32 *newUID,
                       int32 uid,
                       AGRecordStatus mod,
                       int32 recordDataLength,
                       void *recordData,
                       int32 platformDataLength,
                       void *platformData)
{
    PalmSyncInfo *pInfo = (PalmSyncInfo *)out;

    debug_print("doCmdAG_RECORD_CMD()\n");

    if(mod == AG_RECORD_NEW_TEMPORARY_UID) {
        uid = 0;
    }

    if (AG_RECORD_DELETED == mod) {

        dlp_DeleteRecord(sd, pInfo->pilot_rHandle, 0, uid);

    } else if (recordDataLength <= 0x0000ffff) {

        if (pInfo->currentDBIsResourceType) {

            dlp_WriteRecord(sd, pInfo->pilot_rHandle, dlpDBFlagResource,
                            uid, 0, (void *)recordData,
                            recordDataLength, (recordid_t *)newUID);
            debug_print("doCmdAG_RECORD_CMD()\n");


        } else {

            dlp_WriteRecord(sd, pInfo->pilot_rHandle, 0,
                            uid, 0, (void *)recordData,
                            recordDataLength, (recordid_t *)newUID);
        }

    }
    return AGCLIENT_CONTINUE;
}
/*----------------------------------------------------------------------------*/
static int32 initAndOpenDatabase(void *_pInfo, AGDBConfig *db,
                                 int32 *errCode)
{
    long result;
    PalmSyncInfo * pInfo = (PalmSyncInfo *)_pInfo;

    if (NULL == db->dbname) {
        *errCode = AGCLIENT_OPEN_ERR;
        return AGCLIENT_ERR;
    }

    result = openDatabase(pInfo, db->dbname, FALSE);
    if (result < 0) {
        if (result == dlpErrNotFound)
            *errCode = AGCLIENT_OPEN_ERR;
        else
            *errCode = AGCLIENT_UNKNOWN_ERR;
        return AGCLIENT_ERR;
    }

    pInfo->pilot_RecIndex = 0;
    pInfo->record = AGRecordNew(0, AG_RECORD_UNMODIFIED, 0, 0, 0, 0);
    if (!pInfo->record) {
        *errCode = AGCLIENT_OPEN_ERR;
        return AGCLIENT_ERR;
    }

    return AGCLIENT_IDLE;
}
/*----------------------------------------------------------------------------*/
static int32 leaveGetRecord(PalmSyncInfo * pInfo, int32 result)
{
    if (pInfo->record) {

        /* Set recordData to NULL so that AGRecordFree doesn't try
         to free it (we own that buffer and will release it when the
         program ends). */
        pInfo->record->recordData = NULL;

        AGRecordFree(pInfo->record);
        pInfo->record = NULL;
    }
    return result;
}
/*----------------------------------------------------------------------------*/
static int32 getRecordBase(PalmSyncInfo * pInfo, AGBool modonly,
                           AGRecord ** record, int32 * errCode)
{
    int32 result;
    int att = 0;
    int cat = 0;
    int size = pInfo->pilot_buffer_size;
    int idx   = pInfo->pilot_RecIndex++;

    result = (modonly) ?
        dlp_ReadNextModifiedRec (sd, pInfo->pilot_rHandle, pInfo->pilot_buffer,
                                 &pInfo->id, &idx,
                                 &size, &att, &cat)
        :
        dlp_ReadRecordByIndex(sd, pInfo->pilot_rHandle, idx,
                              pInfo->pilot_buffer, &pInfo->id,
                              &size, &att, &cat);

    if (result < 0) {
        closeDatabase(pInfo);
        if (result == dlpErrNotFound) {
            debug_print("(successfully reached end of records ...)\n");
            return leaveGetRecord(pInfo, AGCLIENT_IDLE);
        }
        else {
            *errCode = AGCLIENT_UNKNOWN_ERR;
            return leaveGetRecord(pInfo, AGCLIENT_ERR);
        }
    }
    pInfo->record = AGRecordInit(pInfo->record, pInfo->id,
                                 AGPalmPilotAttribsToMALMod((uint8)att),
                                 size, pInfo->pilot_buffer, 0, NULL);

    *record = pInfo->record;
    return AGCLIENT_CONTINUE;
}
/*----------------------------------------------------------------------------*/
static int32 getNextModifiedRecord(void * out, AGRecord ** record,
                                   int32 *errCode)
{
    debug_print("GetNextModifiedRecord()\n");

    return getRecordBase((PalmSyncInfo *)out, TRUE, record, errCode);
}
/*----------------------------------------------------------------------------*/
static int32 getNextRecord(void * out, AGRecord ** record, int32 *errCode)
{
    debug_print("GetNextRecord()\n");

    return getRecordBase((PalmSyncInfo *)out, FALSE, record, errCode);
}
/*----------------------------------------------------------------------------*/
static AGBool doClientProcessorLoop(PalmSyncInfo * pInfo, AGNetCtx *ctx)
{
    int32 dummyError;
    int32 cpResult;
    int32 syncCount;
    int32 i, n;
    AGBool cancelled = FALSE;
    AGLocationConfig *lc = NULL;

    n = AGUserConfigCount(pInfo->userConfig);

    for (i = 0; i < n; ++i) {

        AGServerConfig * sc =
            AGUserConfigGetServerByIndex(pInfo->userConfig, i);

        if (cancelled)
            continue;

        if (NULL == sc)
            continue;

        if (sc->disabled)
            continue;

        if (NULL == sc->serverName || sc->serverPort <= 0)
            continue;

        syncCount = 0;
        doStartServer(pInfo, sc, &dummyError);

        do {
            AGCommandProcessorStart(pInfo->commandProcessor);

            pInfo->deviceInfo = AGDeviceInfoNew();
            /* If this fails, we're totally dead.  Exit with fail code. */
            if(!pInfo->deviceInfo)
                return FALSE;

            readDeviceInfo(pInfo);

            if (httpProxy && httpProxyPort) {
                lc = AGLocationConfigNew();
                lc->HTTPUseProxy = 1;
                lc->HTTPName = httpProxy;
                lc->HTTPPort = httpProxyPort;
                jp_logf (LOG_INFO | LOG_GUI, "Using Proxy Server: Address %s: Port %d\n", httpProxy,
                       httpProxyPort);                }
            if (proxyUsername && proxyPassword) {
                lc->HTTPUseAuthentication = 1;
                lc->HTTPUsername = proxyUsername;
                lc->HTTPPassword = proxyPassword;
            }

            jp_logf (LOG_INFO | LOG_GUI, "Connecting to %s: %s\n", sc->friendlyName, sc->serverName);
            if (socksProxy && socksProxyPort) {
                if (!lc)
                    lc = AGLocationConfigNew();
                lc->SOCKSUseProxy = 1;
                lc->SOCKSName = socksProxy;
                lc->SOCKSPort = socksProxyPort;
                jp_logf (LOG_INFO | LOG_GUI, "Using SOCKS proxy: Address %s: Port %d\n", socksProxy,
                       socksProxyPort);
            }



            pInfo->clientProcessor =
                AGClientProcessorNew(pInfo->currentServerConfig,
                                     pInfo->deviceInfo,
                                     lc,
                                     pInfo->platform,
                                     TRUE,
                                     ctx);

            /* If this fails, we're totally dead.  Exit with fail code. */
            if (NULL == pInfo->clientProcessor) {
                AGDeviceInfoFree(pInfo->deviceInfo);
                return FALSE;
            }

            if (secnetpresync) {

                if (sc->connectSecurely && sc->allowSecureConnection)
                    jp_logf (LOG_INFO | LOG_GUI, "Secure Connection\n");

                (*secnetpresync) (ctx, sc, lc,
                                  &pInfo->clientProcessor->syncProcessor,
                                  (sc->connectSecurely &&
                                   sc->allowSecureConnection));
            }

            AGClientProcessorSetBufferServerCommands(pInfo->clientProcessor,
                                                     FALSE);

            AGClientProcessorSync(pInfo->clientProcessor);


            cpResult = AGCLIENT_CONTINUE;
            while (AGCLIENT_CONTINUE == cpResult) {
                cpResult = AGClientProcessorProcess(pInfo->clientProcessor);

                if (AGCLIENT_CONTINUE == cpResult && pInfo->quit) {
                    cancelled = TRUE;
                    cpResult = AGCLIENT_IDLE;
                }
                if (dlp_OpenConduit(sd)<0) {
                    jp_logf (LOG_FATAL, "Exiting on cancel, data not retrieved.\n");
                    exit(1);
                }

            }


            if(cpResult == AGCLIENT_ERR) {

                char *msg = AGGetMsg(pInfo->clientProcessor->errStringId);
                if (msg)
                    jp_logf (LOG_FATAL,"%s\n", msg);
                else
                    jp_logf (LOG_FATAL,"Unknown error\n");

            }

            if (secnetpostsync) {
                (*secnetpostsync) (ctx, sc, lc,
                                   &pInfo->clientProcessor->syncProcessor,
                                   (sc->connectSecurely &&
                                    sc->allowSecureConnection));
            }

            AGClientProcessorFree(pInfo->clientProcessor);
            AGDeviceInfoFree(pInfo->deviceInfo);

        } while (!cancelled
                 && AGCommandProcessorShouldSyncAgain(pInfo->commandProcessor)
                 && syncCount++ < 5);
        doEndServer(pInfo, &dummyError);

        /* It's possible to cancel in the middle of an open database,
         so if that happened, close it. */
        if (pInfo->pilot_rHandle)
            closeDatabase(pInfo);

    }

    return TRUE; /* success */
}
/*----------------------------------------------------------------------------*/
int loadSecLib()
{
    int secureLib = 0;
    void *h;
    char *seclib;

    seclib = getenv("MALSYNC_SECURITYLIB");

    if (!seclib)
        return 0;

#ifdef RTLD_GLOBAL
    h  = dlopen(seclib, RTLD_GLOBAL|RTLD_NOW);
#else
    h  = dlopen(seclib, RTLD_NOW);
#endif
    if (h) {
        secnetinit        = dlsym(h, "NetInit");
        secnetclose       = dlsym(h, "NetClose");
        secctxsize        = dlsym(h, "NetGetCtxSize");
        secnetpostsync        = dlsym(h, "NetPostSyncHook");
        secnetpresync        = dlsym(h, "NetPreSyncHook");

        if (secnetinit && secnetclose && secctxsize ) {
            secureLib = 1;
        }
    } else {
        jp_logf (LOG_WARN, "%s\n", dlerror());
    }

    return secureLib;
}
